@@ -3,6 +3,7 @@ import mmap
 import os
 import shutil
 import struct
+import time
 from binascii import hexlify, unhexlify
 from collections import defaultdict
 from configparser import ConfigParser
@@ -10,8 +11,6 @@ from datetime import datetime
 from functools import partial
 from itertools import islice
 
-import msgpack
-
 from .constants import *  # NOQA
 from .hashindex import NSIndex
 from .helpers import Error, ErrorWithTraceback, IntegrityError, format_file_size, parse_file_size
@@ -20,6 +19,7 @@ from .helpers import ProgressIndicatorPercent
 from .helpers import bin_to_hex
 from .helpers import hostname_is_unique
 from .helpers import secure_erase, truncate_and_unlink
+from .helpers import msgpack
 from .locking import Lock, LockError, LockErrorT
 from .logger import create_logger
 from .lrucache import LRUCache
@@ -116,9 +116,18 @@ class Repository:
     class AlreadyExists(Error):
         """A repository already exists at {}."""
 
+    class PathAlreadyExists(Error):
+        """There is already something at {}."""
+
+    class ParentPathDoesNotExist(Error):
+        """The parent path of the repo directory [{}] does not exist."""
+
     class InvalidRepository(Error):
         """{} is not a valid repository. Check repo config."""
 
+    class InvalidRepositoryConfig(Error):
+        """{} does not have a valid configuration. Check repo config [{}]."""
+
     class AtticRepository(Error):
         """Attic repository detected. Please run "borg upgrade {}"."""
 
@@ -140,7 +149,8 @@ class Repository:
         """The storage quota ({}) has been exceeded ({}). Try deleting some archives."""
 
     def __init__(self, path, create=False, exclusive=False, lock_wait=None, lock=True,
-                 append_only=False, storage_quota=None, check_segment_magic=True):
+                 append_only=False, storage_quota=None, check_segment_magic=True,
+                 make_parent_dirs=False):
         self.path = os.path.abspath(path)
         self._location = Location('file://%s' % self.path)
         self.io = None  # type: LoggedIO
@@ -161,6 +171,7 @@ class Repository:
         self.storage_quota_use = 0
         self.transaction_doomed = None
         self.check_segment_magic = check_segment_magic
+        self.make_parent_dirs = make_parent_dirs
 
     def __del__(self):
         if self.lock:
@@ -195,9 +206,24 @@ class Repository:
     def id_str(self):
         return bin_to_hex(self.id)
 
+    @staticmethod
+    def is_repository(path):
+        """Check whether there is already a Borg repository at *path*."""
+        try:
+            # Use binary mode to avoid troubles if a README contains some stuff not in our locale
+            with open(os.path.join(path, 'README'), 'rb') as fd:
+                # Read only the first ~100 bytes (if any), in case some README file we stumble upon is large.
+                readme_head = fd.read(100)
+                # The first comparison captures our current variant (REPOSITORY_README), the second comparison
+                # is an older variant of the README file (used by 1.0.x).
+                return b'Borg Backup repository' in readme_head or b'Borg repository' in readme_head
+        except OSError:
+            # Ignore FileNotFound, PermissionError, ...
+            return False
+
     def check_can_create_repository(self, path):
         """
-        Raise self.AlreadyExists if a repository already exists at *path* or any parent directory.
+        Raise an exception if a repository already exists at *path* or any parent directory.
 
         Checking parent directories is done for two reasons:
         (1) It's just a weird thing to do, and usually not intended. A Borg using the "parent" repository
@@ -207,8 +233,11 @@ class Repository:
             repository, user's can only use the quota'd repository, when their --restrict-to-path points
             at the user's repository.
         """
-        if os.path.exists(path) and (not os.path.isdir(path) or os.listdir(path)):
-            raise self.AlreadyExists(path)
+        if os.path.exists(path):
+            if self.is_repository(path):
+                raise self.AlreadyExists(path)
+            if not os.path.isdir(path) or os.listdir(path):
+                raise self.PathAlreadyExists(path)
 
         while True:
             # Check all parent directories for Borg's repository README
@@ -218,25 +247,21 @@ class Repository:
             if path == previous_path:
                 # We reached the root of the directory hierarchy (/.. = / and C:\.. = C:\).
                 break
-            try:
-                # Use binary mode to avoid troubles if a README contains some stuff not in our locale
-                with open(os.path.join(path, 'README'), 'rb') as fd:
-                    # Read only the first ~100 bytes (if any), in case some README file we stumble upon is large.
-                    readme_head = fd.read(100)
-                    # The first comparison captures our current variant (REPOSITORY_README), the second comparison
-                    # is an older variant of the README file (used by 1.0.x).
-                    if b'Borg Backup repository' in readme_head or b'Borg repository' in readme_head:
-                        raise self.AlreadyExists(path)
-            except OSError:
-                # Ignore FileNotFound, PermissionError, ...
-                pass
+            if self.is_repository(path):
+                raise self.AlreadyExists(path)
 
     def create(self, path):
         """Create a new empty repository at `path`
         """
         self.check_can_create_repository(path)
+        if self.make_parent_dirs:
+            parent_path = os.path.join(path, os.pardir)
+            os.makedirs(parent_path, exist_ok=True)
         if not os.path.exists(path):
-            os.mkdir(path)
+            try:
+                os.mkdir(path)
+            except FileNotFoundError as err:
+                raise self.ParentPathDoesNotExist(path) from err
         with open(os.path.join(path, 'README'), 'w') as fd:
             fd.write(REPOSITORY_README)
         os.mkdir(os.path.join(path, 'data'))
@@ -267,7 +292,8 @@ class Repository:
                 os.link(config_path, old_config_path)
             except OSError as e:
                 if e.errno in (errno.EMLINK, errno.ENOSYS, errno.EPERM, errno.ENOTSUP):
-                    logger.warning("Hardlink failed, cannot securely erase old config file")
+                    logger.warning("Failed to securely erase old repository config file (hardlinks not supported>). "
+                                   "Old repokey data, if any, might persist on physical storage.")
                 else:
                     raise
 
@@ -367,11 +393,15 @@ class Repository:
         else:
             self.lock = None
         self.config = ConfigParser(interpolation=None)
-        self.config.read(os.path.join(self.path, 'config'))
+        with open(os.path.join(self.path, 'config')) as fd:
+            self.config.read_file(fd)
         if 'repository' not in self.config.sections() or self.config.getint('repository', 'version') != 1:
             self.close()
             raise self.InvalidRepository(path)
         self.max_segment_size = self.config.getint('repository', 'max_segment_size')
+        if self.max_segment_size >= MAX_SEGMENT_SIZE_LIMIT:
+            self.close()
+            raise self.InvalidRepositoryConfig(path, 'max_segment_size >= %d' % MAX_SEGMENT_SIZE_LIMIT)  # issue 3592
         self.segments_per_dir = self.config.getint('repository', 'segments_per_dir')
         self.additional_free_space = parse_file_size(self.config.get('repository', 'additional_free_space', fallback=0))
         # append_only can be set in the constructor
@@ -625,7 +655,7 @@ class Repository:
             logger.warning('Failed to check free space before committing: ' + str(os_error))
             return
         # f_bavail: even as root - don't touch the Federal Block Reserve!
-        free_space = st_vfs.f_bavail * st_vfs.f_bsize
+        free_space = st_vfs.f_bavail * st_vfs.f_frsize
         logger.debug('check_free_space: required bytes {}, free bytes {}'.format(required_free_space, free_space))
         if free_space < required_free_space:
             if self.created:
@@ -944,6 +974,23 @@ class Repository:
             logger.info('Completed repository check, no problems found.')
         return not error_found or repair
 
+    def scan_low_level(self):
+        """Very low level scan over all segment file entries.
+
+        It does NOT care about what's committed and what not.
+        It does NOT care whether an object might be deleted or superceded later.
+        It just yields anything it finds in the segment files.
+
+        This is intended as a last-resort way to get access to all repo contents of damaged repos,
+        when there is uncommitted, but valuable data in there...
+        """
+        for segment, filename in self.io.segment_iterator():
+            try:
+                for tag, key, offset, data in self.io.iter_objects(segment, include_data=True):
+                    yield key, data, tag, segment, offset
+            except IntegrityError as err:
+                logger.error('Segment %d (%s) has IntegrityError(s) [%s] - skipping.' % (segment, filename, str(err)))
+
     def _rollback(self, *, cleanup):
         """
         """
@@ -1117,20 +1164,21 @@ class LoggedIO:
 
     def __init__(self, path, limit, segments_per_dir, capacity=90):
         self.path = path
-        self.fds = LRUCache(capacity,
-                            dispose=self.close_fd)
+        self.fds = LRUCache(capacity, dispose=self._close_fd)
         self.segment = 0
         self.limit = limit
         self.segments_per_dir = segments_per_dir
         self.offset = 0
         self._write_fd = None
+        self._fds_cleaned = 0
 
     def close(self):
         self.close_segment()
         self.fds.clear()
         self.fds = None  # Just to make sure we're disabled
 
-    def close_fd(self, fd):
+    def _close_fd(self, ts_fd):
+        ts, fd = ts_fd
         safe_fadvise(fd.fileno(), 0, 0, 'DONTNEED')
         fd.close()
 
@@ -1244,13 +1292,37 @@ class LoggedIO:
         return self._write_fd
 
     def get_fd(self, segment):
-        try:
-            return self.fds[segment]
-        except KeyError:
+        # note: get_fd() returns a fd with undefined file pointer position,
+        # so callers must always seek() to desired position afterwards.
+        now = time.monotonic()
+
+        def open_fd():
             fd = open(self.segment_filename(segment), 'rb')
-            self.fds[segment] = fd
+            self.fds[segment] = (now, fd)
             return fd
 
+        def clean_old():
+            # we regularly get rid of all old FDs here:
+            if now - self._fds_cleaned > FD_MAX_AGE // 8:
+                self._fds_cleaned = now
+                for k, ts_fd in list(self.fds.items()):
+                    ts, fd = ts_fd
+                    if now - ts > FD_MAX_AGE:
+                        # we do not want to touch long-unused file handles to
+                        # avoid ESTALE issues (e.g. on network filesystems).
+                        del self.fds[k]
+
+        clean_old()
+        try:
+            ts, fd = self.fds[segment]
+        except KeyError:
+            fd = open_fd()
+        else:
+            # we only have fresh enough stuff here.
+            # update the timestamp of the lru cache entry.
+            self.fds.upd(segment, (now, fd))
+        return fd
+
     def close_segment(self):
         # set self._write_fd to None early to guard against reentry from error handling code paths:
         fd, self._write_fd = self._write_fd, None
@@ -1324,23 +1396,34 @@ class LoggedIO:
             del self.fds[segment]
         backup_filename = filename + '.beforerecover'
         os.rename(filename, backup_filename)
+        if os.path.getsize(backup_filename) < MAGIC_LEN + self.header_fmt.size:
+            # this is either a zero-byte file (which would crash mmap() below) or otherwise
+            # just too small to be a valid non-empty segment file, so do a shortcut here:
+            with open(filename, 'wb') as fd:
+                fd.write(MAGIC)
+            return
         with open(backup_filename, 'rb') as backup_fd:
-            # note: file must not be 0 size (windows can't create 0 size mapping)
+            # note: file must not be 0 size or mmap() will crash.
             with mmap.mmap(backup_fd.fileno(), 0, access=mmap.ACCESS_READ) as mm:
+                # memoryview context manager is problematic, see https://bugs.python.org/issue35686
                 data = memoryview(mm)
-                with open(filename, 'wb') as fd:
-                    fd.write(MAGIC)
-                    while len(data) >= self.header_fmt.size:
-                        crc, size, tag = self.header_fmt.unpack(data[:self.header_fmt.size])
-                        if size < self.header_fmt.size or size > len(data):
-                            data = data[1:]
-                            continue
-                        if crc32(data[4:size]) & 0xffffffff != crc:
-                            data = data[1:]
-                            continue
-                        fd.write(data[:size])
-                        data = data[size:]
-                data.release()
+                d = data
+                try:
+                    with open(filename, 'wb') as fd:
+                        fd.write(MAGIC)
+                        while len(d) >= self.header_fmt.size:
+                            crc, size, tag = self.header_fmt.unpack(d[:self.header_fmt.size])
+                            if size < self.header_fmt.size or size > len(d):
+                                d = d[1:]
+                                continue
+                            if crc32(d[4:size]) & 0xffffffff != crc:
+                                d = d[1:]
+                                continue
+                            fd.write(d[:size])
+                            d = d[size:]
+                finally:
+                    del d
+                    data.release()
 
     def read(self, segment, offset, id, read_data=True):
         """
